#pragma once
#include "tasks/optimization_task.h"

#include "tasks/accuracy/accuracy.h"
#include "tasks/accuracy/cost_complex_accuracy.h"
#include "tasks/accuracy/balanced_accuracy.h"

#include "tasks/regression/regression.h"
#include "tasks/regression/cost_complex_regression.h"
#include "tasks/regression/piecewise_linear_regression.h"
#include "tasks/regression/piecewise_simple_linear_regression.h"


#include "tasks/cost_sensitive.h"
#include "tasks/instance_cost_sensitive.h"
#include "tasks/f1score.h"
#include "tasks/group_fairness.h"
#include "tasks/eq_opp.h"
#include "tasks/prescriptive_policy.h"
#include "tasks/survival_analysis.h"